(* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gont.ml,v 1.5 2002/02/18 11:25:26 malekith Exp $
 *
 * Module implementation: Gont
 * Description: Top level module for Gont compiler. There are two main entry
 * points here. First one, 'compile' compiles specified source file, the 
 * second 'compile_iface' compiles interface. 
 *
 * Compilation of source file involves:
 *  - initialization of Ksi code generator
 *  - opening input and output files
 *  - parsing it -- we parse one declaration at a time, and generate code
 *    for it. This generally involves sequence of:
 *      - Gontbind.bind -- binds names to symbols, reports any 
 *        undefined symbols
 *      - Gonttypes.typecheck_decl -- checks whatever declaration is 
 *        typecorrect
 *      - Gontcodegen.ksi_code -- returns list of Ksi declarations to
 *        be output (we might generate more then one Ksi declarations
 *        from given Gont declaration, hence list).
 *  - closing input and output files
 * 
 * Compilation of interface, is at the very moment simplified. It does
 * not check whatever given declaration has already been seen (we should
 * report error then), and few other things. We generally:
 *  - open input file
 *  - parse it, this results in list of declarations
 *  - pass it to Gontmodule.save function, that makes use of OCaml 
 *    serialization code.
 *  - close input file
 *)

type compilation_info = {ci_input_file: string;
                         ci_output_file: string;
                         ci_target_prefix: string;
                         ci_module_name: string}
open Gonttree
open Gontutil
 
let open_files info =
  let in_ch = 
    if info.ci_input_file = "-" then 
       stdin 
     else 
       open_in info.ci_input_file in
  let out_ch = 
    if info.ci_output_file = "-" then 
       stdout
     else 
       open_out info.ci_output_file in
  (in_ch, out_ch)
let close_files (in_ch, out_ch) =
     if out_ch != stdout then close_out out_ch else ();
     if in_ch != stdin then close_in in_ch else ()
let loc (f, l, c) = {loc_file = f; loc_line = l; loc_char = c}
;;

let do_compile iface info =
  let (in_ch, out_ch) = open_files info in
  try
    Gontutil.reset ();
    Gontmodule.reset ();
    let bind_ctx = ref (Gontbind.init_bind info.ci_module_name iface) in
    let codegen_ctx = 
      if not iface then
         ref (Some (Gontcodegen.init_codegen info.ci_module_name
                                             info.ci_input_file))
      else
         ref None in
    let compile_decl decl =
      bind_ctx := Gontbind.bind !bind_ctx decl;
        Gonttypes.typecheck_decl decl;
        match !codegen_ctx with
        | Some ctx -> 
          let (ctx', code) = Gontcodegen.ksi_code ctx decl in
          codegen_ctx := Some ctx'; code 
        | None -> assert false in
    let compile_iface_decl decl =
      bind_ctx := Gontbind.bind !bind_ctx decl;
      Gonttypes.typecheck_decl decl in
    let lexbuf = Lexing.from_channel in_ch in
    if not iface then (
      Ksi.read_target_conf info.ci_target_prefix;
      Ksi.output_header out_ch;
      match !codegen_ctx with
      | Some ctx -> 
        List.iter (Ksi.output_decl out_ch) 
                  (Gontcodegen.ksi_runtime_proto ctx)
      | None -> assert false);
    Lexer.set_file_name info.ci_input_file;
    (if iface then
      (if info.ci_module_name <> "Std" &&
          info.ci_module_name <> "Stdc" then
         (compile_iface_decl (TopLevelOpen "Stdc");
          compile_iface_decl (TopLevelOpen "Std")))
     else (
       List.iter (Ksi.output_decl out_ch) 
                 (compile_decl (TopLevelOpen "Stdc"));
       List.iter (Ksi.output_decl out_ch) 
                 (compile_decl (TopLevelOpen "Std"))));
    (try
      while true do
        let decl = 
          try
            if iface then
              (Parser.iface_decl Lexer.token lexbuf) 
            else
              (Parser.decl Lexer.token lexbuf) 
          with 
              Parsing.Parse_error ->
                raise(Syntax_error (loc (Lexer.get_current_location ()),
                                    "parse error"))
            | Lexer.Eof -> 
                raise Eof
            | Lexer.Error(f, l, c, e) ->
                raise(Syntax_error (loc (f, l, c), e))
        in
        if iface then
          compile_iface_decl decl
        else
          List.iter (Ksi.output_decl out_ch) (compile_decl decl)
      done
    with Eof -> 
      if iface then
        Gontbind.finalize_module out_ch !bind_ctx
      else
        ());
    if not iface then (
      (match !codegen_ctx with
       | Some ctx -> Gontcodegen.check_module ctx;
       | None -> assert false);
      Ksi.output_trailer out_ch);
    close_files (in_ch, out_ch);
    0
  with 
      Syntax_error(loc, msg) ->
        close_files (in_ch, out_ch);
        err_msg (loc, msg);
        1
    | Error(msg) ->
        close_files (in_ch, out_ch);
        err_msg ({loc_file=info.ci_input_file; loc_line=0; loc_char=0}, msg);
        1
    | x ->
        close_files (in_ch, out_ch);
        prerr_endline ("Unhandled exception: " ^ (Printexc.to_string x));
        1
;;

let compile = do_compile false
let compile_iface = do_compile true
;;

let dump_module m =
  print_endline (xf "// module dump of %s" m);
  let (g, t) = Gontmodule.globals m in
  List.iter (fun t -> 
                print_endline (xf "type %s = %s;" t.td_name
                                  (Gonttypes.string_of_type t.td_type)))
           t;
  List.iter (fun sym -> 
                print_endline (xf "%s %s;" (Gonttypes.string_of_type
                               sym.sym_type) sym.sym_name))
           g;
  0
;;
